﻿using SimpleJSON;
using System.Collections.Generic;
using System.Linq;
using ToySerialController.UI;
using ToySerialController.Utils;
using UnityEngine;
using DebugUtils;
using System;
using Leap.Unity.Infix;
using Leap.Unity;

namespace ToySerialController.MotionSource
{
    public abstract class AbstractFemaleMotionSource : AbstractRefreshableMotionSource
    {
        private Atom _femaleAtom;
        private Vector3 _targetPosition;
        private Vector3 _targetUp;
        private Vector3 _targetRight;
        private Vector3 _targetForward;

        private float _referenceLengthMultiplier = 1f;
        private float _referenceRadiusMultiplier = 1f;

        protected string _invalidAutoTargetUid = "";

        private JSONStorableStringChooser FemaleChooser;
        private JSONStorableStringChooser TargetChooser;
        private JSONStorableBool IncludeMales;

        private SuperController Controller => SuperController.singleton;

        public override Vector3 TargetPosition => _targetPosition;
        public override Vector3 TargetUp => _targetUp;
        public override Vector3 TargetRight => _targetRight;
        public override Vector3 TargetForward => _targetForward;

        public override float ReferenceLengthMultiplier { get { return _referenceLengthMultiplier; } }
        public override float ReferenceRadiusMultiplier { get { return _referenceRadiusMultiplier; } }

        public override void CreateUI(IUIBuilder builder)
        {
            var targets = new List<string> { "Auto", "Vagina", "Pelvis", "Hips", "Anus", "Mouth", "Left Hand", "Right Hand", "Chest", "ChestExtended", "Left Foot", "Right Foot", "Feet" };
            var defaultTarget = "Auto";

            FemaleChooser = builder.CreatePopup("MotionSource:Female", "Select Female", null, null, FemaleChooserCallback);
            TargetChooser = builder.CreateScrollablePopup("MotionSource:FemaleTarget", "Select Target Point", targets, defaultTarget, null);
            IncludeMales = builder.CreateToggle($"MotionSource:IncludeMales", "Include Males In Female Targets", false, (b) => { RefreshButtonCallback(); });

            FindFemales();

            base.CreateUI(builder);
        }

        public override void DestroyUI(IUIBuilder builder)
        {
            builder.Destroy(FemaleChooser);
            builder.Destroy(TargetChooser);
            builder.Destroy(IncludeMales);

            base.DestroyUI(builder);
        }

        public override void StoreConfig(JSONNode config)
        {
            config.Store(IncludeMales);
            config.Store(FemaleChooser);
            config.Store(TargetChooser);
        }

        public override void RestoreConfig(JSONNode config)
        {
            config.Restore(IncludeMales);
            config.Restore(FemaleChooser);
            config.Restore(TargetChooser);

            FindFemales();
        }

        public override bool Update(Func<IMotionSource, bool> motionTestFunc)
        {
            if (FemaleChooser.val == FEMALE_AUTO)
            {
                Atom bestFemale = null;
                Vector3 bestTargetPosition = _targetPosition;
                Vector3 bestTargetUp = _targetUp;
                Vector3 bestTargetRight = _targetRight;
                Vector3 bestTargetForward = _targetForward;
                float bestReferenceLengthMultiplier = _referenceLengthMultiplier;
                float bestReferenceRadiusMultiplier = _referenceRadiusMultiplier;
                float bestMotionDistance = float.MaxValue;
                int firstIndexTest = 0;
                if (_femaleAtom != null)
                {
                    firstIndexTest = _femaleUids.IndexOf(_femaleAtom.uid);
                    if (firstIndexTest == -1)
                        firstIndexTest = 0;
                }

                for (int i = 0; i < _femaleUids.Count; i++)
                {
                    if (i == 0)
                        _femaleAtom = Controller.GetAtomByUid(_femaleUids[firstIndexTest]);
                    else if (i == firstIndexTest)
                        _femaleAtom = Controller.GetAtomByUid(_femaleUids[0]);
                    else
                        _femaleAtom = Controller.GetAtomByUid(_femaleUids[i]);

                    float motionDistance = 0f;
                    if (UpdateCurrentFemale(motionTestFunc, out motionDistance) && bestMotionDistance > motionDistance)
                    {
                        bestMotionDistance = motionDistance;
                        bestTargetPosition = _targetPosition;
                        bestTargetUp = _targetUp;
                        bestTargetRight = _targetRight;
                        bestTargetForward = _targetForward;
                        bestFemale = _femaleAtom;
                        bestReferenceLengthMultiplier = _referenceLengthMultiplier;
                        bestReferenceRadiusMultiplier = _referenceRadiusMultiplier;
                    }
                }

                if (bestFemale != null)
                {
                    _femaleAtom = bestFemale;
                    _targetPosition = bestTargetPosition;
                    _targetUp = bestTargetUp;
                    _targetRight = bestTargetRight;
                    _targetForward = bestTargetForward;
                    _referenceLengthMultiplier = bestReferenceLengthMultiplier;
                    _referenceRadiusMultiplier = bestReferenceRadiusMultiplier;
                    return true;
                }
                _referenceLengthMultiplier = 1f;
                _referenceRadiusMultiplier = 1f;
            }
            else
            {
                float motionDistance = 0f;
                return UpdateCurrentFemale(motionTestFunc, out motionDistance);
            }

            return false;
        }

        private bool UpdateCurrentFemale(Func<IMotionSource, bool> motionTestFunc, out float motionDistance)
        {
            if (_femaleAtom == null || _femaleAtom.uid == _invalidAutoTargetUid)
            {
                motionDistance = float.MaxValue;
                return false;
            }

            if (UpdateTarget(motionTestFunc, out motionDistance))
            {
                DebugDraw.DrawTransform(TargetPosition, TargetUp, TargetRight, TargetForward, 0.15f);
                return true;
            }

            return false;
        }

        private bool UpdateTarget(Func<IMotionSource, bool> motionTestFunc, out float motionDistance)
        {
            bool result = false;

            if (TargetChooser.val == "Auto") return UpdateAutoTarget(motionTestFunc, out motionDistance);
            else if (TargetChooser.val == "Vagina") result = UpdateVaginaTarget();
            else if (TargetChooser.val == "Pelvis") result = UpdateFreeControllerTarget("pelvisControl");
            else if (TargetChooser.val == "Hips") result = UpdateFreeControllerTarget("hipControl");
            else if (TargetChooser.val == "Anus") result = UpdateAnusTarget();
            else if (TargetChooser.val == "Mouth") result = UpdateMouthTarget();
            else if (TargetChooser.val == "Left Hand") result = UpdateLeftHandTarget();
            else if (TargetChooser.val == "Right Hand") result = UpdateRightHandTarget();
            else if (TargetChooser.val == "Chest") result = UpdateChestTarget();
            else if (TargetChooser.val == "ChestExtended") result = UpdateChestExtendedTarget();
            else if (TargetChooser.val == "Left Foot") result = UpdateLeftFootTarget();
            else if (TargetChooser.val == "Right Foot") result = UpdateRightFootTarget();
            else if (TargetChooser.val == "Feet") result = UpdateFeetTarget();

            motionDistance = Vector3.Distance(ReferencePosition, TargetPosition);

            return result;
        }

        private List<Func<bool>> _targets = null;
        private List<string> _targetsInfo = null;
        private int _prevAutoTargetIndex = -1;

        private bool UpdateAutoTarget(Func<IMotionSource, bool> motionTestFunc, out float motionDistance)
        {
            if (_targets == null)
            {
                _targets = new List<Func<bool>>
                {
                    UpdateVaginaTarget,
                    UpdateAnusTarget,
                    UpdateMouthTarget,
                    UpdateLeftHandTarget,
                    UpdateRightHandTarget,
                    UpdateChestExtendedTarget,
                    UpdateFeetTarget,
                    UpdateLeftFootTarget,
                    UpdateRightFootTarget
                };
                _targetsInfo = new List<string> 
                { 
                    "Vagina",
                    "Anal",
                    "Mouth",
                    "LeftHand",
                    "RightHand",
                    "ChestExtended",
                    "Feet",
                    "LeftFoot",
                    "RightFoot"
                };
            }

            motionDistance = float.MaxValue;

            Func<bool> bestPick = null;
            float bestDistance = float.MaxValue;
            int bestIndex = -1;

            Func<bool> bestRunningPick = null;
            float bestRunningPickD = 0f;
            int bestRunningPickIndex = 0;

            for (int i = 0; i < _targets.Count; i ++)
            {
                var target = _targets[i];
                if (target())
                {
                    var distance = Vector3.Distance(ReferencePosition, TargetPosition);
                    if (distance < bestDistance)
                    {
                        bestPick = target;
                        bestDistance = distance;
                        bestIndex = i;
                    }
                    if (motionTestFunc != null && motionTestFunc(this))
                    {
                        float d = Vector3.Dot(TargetUp, ReferenceUp);
                        if (bestRunningPickD < d)
                        {
                            bestRunningPickD = d;
                            bestRunningPick = target;
                            bestRunningPickIndex = i;
                        }
                    }
                }
            }

            if (bestRunningPick != null)
            {
                bestPick = bestRunningPick;
                bestIndex = bestRunningPickIndex;
                bestDistance = -1f; // inside and tracking
            }

            if (_prevAutoTargetIndex != bestIndex && bestIndex >= 0)
            {
                _prevAutoTargetIndex = bestIndex;
                // SuperController.LogMessage(_targetsInfo[bestIndex] + (bestRunningPick != null ? " [Driving Output]" : " *"));
            }

            motionDistance = bestDistance;

            if (bestPick == null)
                return false;

            return bestPick();
        }

        private bool UpdateVaginaTargetMale()
        {
            // Vaginal for male will be Penis
            var penisColliders = new List<string> {
                "AutoColliderGen1Hard",
                "AutoColliderGen2Hard",
                "AutoColliderGen3aHard",
                "AutoColliderGen3bHard"
            };

            var penisTransforms = _femaleAtom
                .GetComponentsInChildren<Collider>()
                .Where(c => penisColliders.Contains(c.name))
                .OrderBy(c => c.name, StringComparer.OrdinalIgnoreCase)
                .Select(c => c.transform)
                .ToList();

            if (penisTransforms.Count != 4)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1.1f;

            var penisLength = 0.0f;
            for (int i = 0, j = 1; j < penisTransforms.Count; i = j++)
                penisLength += Vector3.Distance(penisTransforms[i].position, penisTransforms[j].position);

            var penisOrigin = penisTransforms[0];

            var positionOffset = -penisOrigin.up * (penisTransforms[1].position - penisTransforms[0].position).magnitude * 0.15f;
            var referencePosition = penisOrigin.position + positionOffset;

            _targetUp = (referencePosition - penisOrigin.position).normalized;
            _targetRight = penisTransforms[0].forward;
            _targetForward = penisTransforms[0].right; // Vector3.Cross(_targetRight, _targetForward);

            _targetPosition = penisOrigin.position - _targetUp * (penisLength - 0.005f);

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            return true;
        }

        private bool UpdateVaginaTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateVaginaTargetMale();
            }

            var labiaTrigger = _femaleAtom.GetRigidBodyByName("LabiaTrigger");
            var vaginaTrigger = _femaleAtom.GetRigidBodyByName("VaginaTrigger");
            var positionOffsetCollider = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointB");

            if (labiaTrigger == null || vaginaTrigger == null || positionOffsetCollider == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            _targetPosition = labiaTrigger.transform.position;
            _targetUp = (vaginaTrigger.transform.position - labiaTrigger.transform.position).normalized;
            _targetRight = vaginaTrigger.transform.right;
            _targetForward = Vector3.Cross(_targetRight, _targetUp);

            var positionOffset = positionOffsetCollider.transform.position
                + positionOffsetCollider.transform.forward * positionOffsetCollider.radius
                - _targetUp * 0.0025f;
            _targetPosition += _targetUp * Vector3.Dot(positionOffset - _targetPosition, _targetUp);

            return true;
        }

        private bool UpdateFreeControllerTarget(string id)
        {
            var control = _femaleAtom.GetStorableByID(id) as FreeControllerV3;
            if (control == null)
                return false;

            var followBody = control.followWhenOffRB;
            var labiaTrigger = _femaleAtom.GetRigidBodyByName("LabiaTrigger");
            var vaginaTrigger = _femaleAtom.GetRigidBodyByName("VaginaTrigger");
            var positionOffsetCollider = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointB");

            if (followBody == null || labiaTrigger == null || vaginaTrigger == null || positionOffsetCollider == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            var vaginaPosition = labiaTrigger.transform.position;
            var vaginaUp = (vaginaTrigger.transform.position - labiaTrigger.transform.position).normalized;
            var vaginaRight = vaginaTrigger.transform.right;
            var vaginaForward = Vector3.Cross(vaginaRight, vaginaUp);

            var positionOffset = positionOffsetCollider.transform.position
                + positionOffsetCollider.transform.forward * positionOffsetCollider.radius
                - vaginaUp * 0.0025f;
            vaginaPosition += vaginaUp * Vector3.Dot(positionOffset - vaginaPosition, vaginaUp);

            var controlToBody = control.control.position - followBody.position;
            var referenceToVagina = vaginaPosition - ReferencePosition;

            var controlToBodyRotation = Quaternion.Slerp(control.control.rotation, followBody.rotation, 0.5f).ToNormalized();
            var vaginaRotation = Quaternion.LookRotation(vaginaForward, vaginaUp);
            var rotation = Quaternion.Slerp(vaginaRotation, controlToBodyRotation, 0.125f).ToNormalized();

            _targetUp = rotation.GetUp();
            _targetRight = rotation.GetRight();
            _targetForward = rotation.GetForward();

            var maxOffsetDistance = Mathf.Min(0.05f, referenceToVagina.magnitude);
            var maxControlDistance = 0.25f;

            var controlToBodyUp = Vector3.Project(controlToBody, vaginaUp);
            var controlT = Mathf.Clamp01(controlToBodyUp.magnitude / maxControlDistance);
            var strength = controlT < 0.5 ? 2 * controlT * controlT : -1 + (4 - 2 * controlT) * controlT;
            var sign = -Mathf.Sign(Vector3.Dot(vaginaUp, controlToBodyUp));
            var offsetN = referenceToVagina.normalized;
            var offset = offsetN * sign * maxOffsetDistance * strength;

            _targetPosition = vaginaPosition - offset;
            return true;
        }

        private bool UpdateAnusTargetMale()
        {
            CapsuleCollider anusLeft = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointAlMale");
            CapsuleCollider anusRight = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointArMale");

            if (anusLeft == null || anusRight == null)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1.75f; // for better targeting from behind

            _targetUp = ((anusLeft.transform.up + anusRight.transform.up) / 2).normalized;
            _targetRight = ((anusLeft.transform.right + anusRight.transform.right) / 2).normalized;
            _targetForward = ((anusLeft.transform.forward + anusRight.transform.forward) / 2).normalized;
            _targetPosition = (anusLeft.transform.position + anusRight.transform.position) / 2;

            return true;
        }

        private bool UpdateAnusTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateAnusTargetMale();
            }

            var anusLeft = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointAl");
            var anusRight = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointAr");

            if (anusLeft == null || anusRight == null)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1f;

            _targetUp = ((anusLeft.transform.up + anusRight.transform.up) / 2).normalized;
            _targetRight = ((anusLeft.transform.right + anusRight.transform.right) / 2).normalized;
            _targetForward = ((anusLeft.transform.forward + anusRight.transform.forward) / 2).normalized;
            _targetPosition = (anusLeft.transform.position + anusRight.transform.position) / 2;

            return true;
        }

        private bool UpdateMouthTarget()
        {
            var bottomLip = _femaleAtom.GetComponentByName<Transform>("lowerJawStandardColliders")?.GetComponentByName<CapsuleCollider>("_ColliderLipM");
            var topLip = _femaleAtom.GetComponentByName<Transform>("AutoCollidersTongueUpperLip")?.GetComponentByName<Transform>("AutoColliderAutoCollidersFaceCentral2Hard");
            var mouthTrigger = _femaleAtom.GetRigidBodyByName("MouthTrigger");

            if (bottomLip == null || topLip == null || mouthTrigger == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            var center = (topLip.transform.position + bottomLip.transform.position) / 2;
            _targetUp = (mouthTrigger.transform.position - center).normalized;
            _targetRight = mouthTrigger.transform.right;
            _targetForward = Vector3.Cross(_targetUp, _targetRight);
            _targetPosition = center - TargetUp * Vector3.Distance(center, mouthTrigger.transform.position) * 0.2f;

            DebugDraw.DrawCircle(TargetPosition, TargetUp, TargetRight, Color.gray, (topLip.transform.position - bottomLip.transform.position).magnitude / 2);

            return true;
        }

        private bool UpdateLeftHandTarget() => UpdateHandTarget("l");

        private bool UpdateRightHandTarget() => UpdateHandTarget("r");

        private bool UpdateHandTarget(string side)
        {
            var carpal = _femaleAtom.GetRigidBodyByName($"{side}Carpal2");
            var fingerBase = carpal?.GetComponentByName<CapsuleCollider>("_Collider3");
            var fingerTip = _femaleAtom.GetRigidBodyByName($"{side}Pinky3")?.GetComponentInChildren<CapsuleCollider>();

            if (carpal == null || fingerBase == null || fingerTip == null)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1.2f;

            float sign = 1f;
            if (side == "l")
            {
                _targetRight = -fingerBase.transform.up;
                _targetForward = -fingerBase.transform.right;
                sign = -1f;
            }
            else if (side == "r")
            {
                _targetRight = fingerBase.transform.up;
                _targetForward = fingerBase.transform.right;
            }

            var fingerBasePosition = fingerBase.transform.position - sign * fingerBase.transform.right * (fingerBase.height / 2 - fingerBase.radius) - sign * fingerBase.transform.up * fingerBase.radius;
            var fingerTipPosition = fingerTip.transform.position - fingerTip.transform.right * (fingerTip.height / 2 - fingerTip.radius) - fingerTip.transform.up * fingerTip.radius;
            _targetUp = fingerBase.transform.forward;
            _targetPosition = (fingerBasePosition + fingerTipPosition) / 2 + _targetUp * 0.04f;

            // figure out best of 4 directions to track against Reference
            if (Vector3.Dot(-_targetForward, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = -_targetForward;
                _targetForward = tmp;
            }
            else if (Vector3.Dot(_targetForward, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = _targetForward;
                _targetForward = tmp;
            }
            else if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            DebugDraw.DrawLine(fingerBasePosition, fingerTipPosition, Color.gray);

            return true;
        }

        private bool UpdateChestTargetMale()
        {
            CapsuleCollider left = null;
            CapsuleCollider right = null;
            CapsuleCollider chest = null;
            foreach (var comp in _femaleAtom.GetComponentsInChildren<AutoCollider>())
            {
                if (comp.name == "AutoColliderFemaleAutoColliderschest1")
                {
                    chest = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    left = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    right = comp.jointCollider as CapsuleCollider;
                }
            }

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            _targetUp = chest.transform.right;
            _targetRight = chest.transform.up;
            _targetForward = -chest.transform.forward;

            var leftPosition = left.transform.position + _targetForward * 0.02f;
            var rightPosition = right.transform.position - _targetForward * 0.02f;            
            _targetForward = (rightPosition - leftPosition).normalized;
            _targetUp = -Vector3.Cross(_targetForward, _targetRight);

            var chestPosition = chest.transform.position + _targetForward * (chest.radius + 0.2f) - _targetUp * 0.1f;

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f); // + _targetUp * 0.10f;

            return true;
        }

        private bool UpdateChestTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateChestTargetMale();
            }

            var left = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderslNipple1")?.jointCollider as CapsuleCollider;
            var right = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoCollidersrNipple1")?.jointCollider as CapsuleCollider;
            var chest = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderschest2c")?.jointCollider as CapsuleCollider;

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            _targetForward = chest.transform.forward;

            var leftPosition = left.transform.position;
            var rightPosition = right.transform.position;
            var chestPosition = chest.transform.position + _targetForward * chest.radius;

            _targetRight = (rightPosition - leftPosition).normalized;
            _targetUp = Vector3.Cross(_targetForward, _targetRight);

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f); // + _targetUp * 0.10f;

            return true;
        }

        private bool UpdateChestExtendedTargetMale()
        {
            CapsuleCollider left = null;
            CapsuleCollider right = null;
            CapsuleCollider chest = null;
            foreach (var comp in _femaleAtom.GetComponentsInChildren<AutoCollider>())
            {
                if (comp.name == "AutoColliderFemaleAutoColliderschest1")
                {
                    chest = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    left = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    right = comp.jointCollider as CapsuleCollider;
                }
            }

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            _referenceLengthMultiplier = 1.5f;
            _referenceRadiusMultiplier = 1.1f;

            _targetUp = chest.transform.right;
            _targetRight = chest.transform.up;
            _targetForward = -chest.transform.forward;

            var leftPosition = left.transform.position + _targetForward * 0.02f;
            var rightPosition = right.transform.position - _targetForward * 0.02f;
            _targetForward = (rightPosition - leftPosition).normalized;
            _targetUp = -Vector3.Cross(_targetForward, _targetRight);

            var chestPosition = chest.transform.position + _targetForward * (chest.radius + 0.2f) * 0.90f - _targetUp * 0.1f;

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f) + _targetUp * 0.05f; // chestExtended

            return true;
        }

        private bool UpdateChestExtendedTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateChestExtendedTargetMale();
            }

            var left = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderslNipple1")?.jointCollider as CapsuleCollider;
            var right = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoCollidersrNipple1")?.jointCollider as CapsuleCollider;
            var chest = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderschest2c")?.jointCollider as CapsuleCollider;

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            // make the chest area bigger to allow for less precision boobs
            _referenceLengthMultiplier = 1.5f;
            _referenceRadiusMultiplier = 1.1f;

            _targetForward = chest.transform.forward;

            var leftPosition = left.transform.position;
            var rightPosition = right.transform.position;
            var chestPosition = chest.transform.position + _targetForward * chest.radius * 0.90f;

            _targetRight = (rightPosition - leftPosition).normalized;
            _targetUp = Vector3.Cross(_targetForward, _targetRight);

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            // _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f) - _targetUp * 0.22f; // chestLower
            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f) + _targetUp * 0.05f; // chestExtended

            return true;
        }

        private bool UpdateLeftFootTarget() => UpdateFootTarget("l");

        private bool UpdateRightFootTarget() => UpdateFootTarget("r");

        private bool UpdateFootTarget(string side)
        {
            var foot = _femaleAtom.GetRigidBodyByName($"{side}Foot");
            var footBase = foot?.GetComponentByName<CapsuleCollider>("_Collider6");

            if (foot == null || footBase == null)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1.3f;

            _targetRight = footBase.transform.forward;
            _targetForward = -footBase.transform.up;

            if (side == "l")
                _targetUp = footBase.transform.right;
            else if (side == "r")
                _targetUp = -footBase.transform.right;

            _targetPosition = footBase.transform.position + _targetForward * footBase.radius + _targetUp * 0.02f + _targetRight * 0.02f;

            // figure out best of 4 directions to track against Reference
            if (Vector3.Dot(-_targetRight, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = -_targetRight;
                _targetRight = tmp;
            }
            else if (Vector3.Dot(_targetRight, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = _targetRight;
                _targetRight = tmp;
            }
            else if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            return true;
        }

        private bool UpdateFeetTarget()
        {
            var leftFoot = _femaleAtom.GetRigidBodyByName("lFoot");
            var rightFoot = _femaleAtom.GetRigidBodyByName("rFoot");
            var leftFootBase = leftFoot?.GetComponentByName<CapsuleCollider>("_Collider6");
            var rightFootBase = rightFoot?.GetComponentByName<CapsuleCollider>("_Collider6");

            if (leftFoot == null || rightFoot == null || leftFootBase == null || rightFootBase == null)
                return false;

            var leftPosition = leftFootBase.transform.position - leftFootBase.transform.up * leftFootBase.radius;
            var rightPosition = rightFootBase.transform.position - rightFootBase.transform.up * rightFootBase.radius;

            float dia = (leftPosition - rightPosition).magnitude;
            if (dia > 0.35f) // arbitary magic value before feet are too far away for this tracker to make sense
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1.2f;

            _targetPosition = (leftPosition + rightPosition) / 2;
            _targetForward = ((leftFootBase.transform.forward + rightFootBase.transform.forward) / 2).normalized;
            _targetUp = Vector3.Cross((leftPosition - rightPosition).normalized, _targetForward).normalized;
            _targetRight = Vector3.Cross(_targetUp, _targetForward).normalized;

            // flip _targetUp based on ReferenceUp for a smoother tracking
            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            DebugDraw.DrawCircle(_targetPosition, Quaternion.FromToRotation(Vector3.up, _targetUp), Color.white, dia / 2);

            return true;
        }

        public const string FEMALE_NONE = "None";
        public const string FEMALE_AUTO = "*Auto*";

        private List<string> _femaleUids = new List<string>();

        private void FindFemales()
        {
            var defaultFemale = FemaleChooser.val;

            var people = Controller.GetAtoms().Where(a => a.type == "Person");
            _femaleUids = people
                .Where(a => a.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Female)
                .Select(a => a.uid)
                .ToList();

            if (IncludeMales.val)
            {
                _femaleUids.AddRange(people
                    .Where(a => a.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
                    .Select(a => a.uid));
            }

            if (defaultFemale != FEMALE_AUTO && !_femaleUids.Contains(defaultFemale))
                defaultFemale = FEMALE_AUTO; //  _femaleUids.FirstOrDefault(uid => uid == _femaleAtom?.uid) ?? _femaleUids.FirstOrDefault() ?? FEMALE_NONE;

            List<string> femaleOptions = new List<string>(_femaleUids);
            femaleOptions.Insert(0, FEMALE_AUTO);
            femaleOptions.Insert(0, FEMALE_NONE);

            FemaleChooser.choices = femaleOptions;
            FemaleChooserCallback(defaultFemale);
        }

        protected void FemaleChooserCallback(string s)
        {
            _femaleAtom = Controller.GetAtomByUid(s);
            FemaleChooser.valNoCallback = (_femaleAtom == null ? (s == FEMALE_AUTO ? FEMALE_AUTO : FEMALE_NONE) : s);
        }

        protected override void RefreshButtonCallback() => FindFemales();
    }
}